/*
One of those Weeks Episode 45: Toronto?
Copyright (c) 2008,2012 Blazing Games Inc.
Code written by Billy D. Spelchan

This program is free software; you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation; either version 2 of the License, or (at your option) any later 
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc.,
59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package com.blazinggames.ootw.ep45
{
	// Adobe Imports
	import flash.display.Bitmap;
	import flash.display.BitmapData;
	import flash.display.MovieClip;
	import flash.geom.Rectangle;

	/**
	 * This is a simple raycaster used for rendering a 3D view of a 2D tiled
	 * map.
	 * 
	 * @author Billy D. Spelchan
	 */
	public class MapRenderer
	extends MovieClip
	{
// *********************
// ***** CONSTANTS *****
// *********************

// *********************
// ***** VARIABLES *****
// *********************

		public var _displayWidth:int;
		public var _displayHeight:int;
		protected var _slicesInCircle:int;
		protected var _sliceSize:Number;
		
		protected var _castTileX:Array;
		protected var _castTileY:Array;
		protected var _castDistance:Array;
		protected var _castHeightTable:Array
		protected var _castHitX:Array;
		protected var _castHitY:Array;
		protected var _castTexture:Array;
		protected var _castSlice:Array;
		
		// for map management
		private var _map:Array;
		private var _mapsize:int;
		
		// bitmap rendering
		protected var _rect:Rectangle;
		protected var _bitmapData:BitmapData;
		protected var _bitmap:Bitmap;
		protected var _wallData:BitmapData;
		protected var _ladderData:BitmapData;
		
// ***********************
// ***** CONSTRUCTOR *****
// ***********************

		public function MapRenderer()
		{
			_slicesInCircle = 3200;//2560;
			_sliceSize = 2.0 * Math.PI / Number(_slicesInCircle);
			_displayWidth = 640;
			_castTileX = new Array(_displayWidth);
			_castTileY = new Array(_displayWidth);
			_castHitX = new Array(_displayWidth);
			_castHitY = new Array(_displayWidth);
			_castDistance = new Array(_displayWidth);
			_castHeightTable = new Array(_displayWidth);
			_castTexture = new Array(_displayWidth);
			_castSlice = new Array(_displayWidth);
			
			// build height distortion compensation table
			var angle:Number = -_displayWidth / 2 * _sliceSize;
			for (var cntr:int = 0; cntr < _displayWidth; ++cntr) {
				_castHeightTable[cntr] = 400 / Math.cos(angle);
				angle += _sliceSize;
			}
			
			_displayHeight = 480;
			_rect = new Rectangle(0,0, _displayWidth, _displayHeight);
			_bitmapData = new BitmapData(_displayWidth, _displayHeight);
			_bitmap = new Bitmap(_bitmapData);
			addChild(_bitmap);
			
			var texture:WallTexture = new WallTexture();
			_wallData = new BitmapData(texture.width, texture.height);
			_wallData.draw(texture);
			var ladderTexture:LadderTexture = new LadderTexture();
			_ladderData = new BitmapData(ladderTexture.width, ladderTexture.height);
			_ladderData.draw(ladderTexture);
		}
		
// ***** Map Management *****

		/**
		 * Sets the map to be rendered. The map is stored in a 1D array with
		 * the width and height being equal. The map data is stored in a 
		 * consecutive row format.
		 * 
		 * @param map Array (1D) holding the map data.
		 * @apram mapsize the width/height of the map (maps are square)
		 */
		public function setMap(map:Array, mapsize:int) 
		{			
			_map = map;
			_mapsize = mapsize;
		}
		
// **********************************
// ***** Functions - Raycasting *****
// **********************************

		/**
		 * Performs the raycasting storing the relevant slice data in internal
		 * arrays that other methods can use to process the visible map. This
		 * allows for a number of ways of rendering the map such as solid (used
		 * in ootw 31, textured, and distance tesing.
		 * 
		 * @param playerX player x location on map
		 * @param playerY player y location on map
		 * @param playerD Direction the player is facing
		 */
		protected function performCast(
				playerX:Number,
				playerY:Number,
				playerD:Number)
		{
			var degreesToSlices:Number = playerD * _slicesInCircle / 360
			var angle:Number = (degreesToSlices - _displayWidth / 2) * _sliceSize;
			var startTileX:int = int(playerX);
			var startTileY:int = int(playerY);
			var tileX:int, tileY:int, tileIndex;
			var tileStepX:int, tileStepY:int;
			var stepX:Number, stepY:Number;
			var stepSizeX:Number, stepSizeY:Number;
			var distanceX:Number, distanceY:Number, distance:Number;
			
			for (var cntr:int = 0; cntr < _displayWidth; ++cntr)
			{
				tileX = startTileX;
				tileY = startTileY;
				tileIndex = startTileY * _mapsize + startTileX;
				
				// first, find the size of each step and the distance to the first step
				stepX = Math.sin(angle);
				if (stepX == 0) {
					stepSizeX = 10000000; // big number
					tileStepX = 1; // shouldn't matter
					distanceX = stepSizeX;
				} else if (stepX < 0) {
					stepSizeX = -1/stepX;
					tileStepX = -1;
					distanceX = (playerX - Number(startTileX)) * stepSizeX;
				} else {
					stepSizeX = 1/stepX;
					tileStepX = 1;
					distanceX = (Number(startTileX+1) - playerX) * stepSizeX;
				}
				stepY = -Math.cos(angle);
				if (stepY == 0) {
					stepSizeY = 10000000; // big number
					tileStepY = -1;
					distanceY = stepSizeY;
				} else if (stepY < 0) {
					stepSizeY = -1/stepY;
					tileStepY = -1;
					distanceY = (playerY - Number(startTileY)) * stepSizeY;
				} else {
					stepSizeY = 1/stepY;
					tileStepY = 1; // y is the big step
					distanceY = (Number(startTileY+1) - playerY) * stepSizeY;
				}
				
				// now keep stepping through tiles until we hit a wall
				var wallHitX:Boolean;
				do {
					if (distanceX < distanceY) {
						tileX += tileStepX;
						tileIndex += tileStepX;
						distance = distanceX;
						distanceX += stepSizeX;
						wallHitX = false;
					} else {
						tileY += tileStepY;
						tileIndex += (tileStepY * _mapsize);
						distance = distanceY;
						distanceY += stepSizeY;
						wallHitX = true;
					}
				} while (_map[tileIndex] < 1);
				if (_map[tileIndex] < 3) // ladder
					_castTexture[cntr] = _ladderData;
				else // wall
					_castTexture[cntr] = _wallData
				_castTileX[cntr] = tileX;
				_castTileY[cntr] = tileY;
				_castDistance[cntr] = distance;
				
				_castHitX[cntr] = playerX + distance * stepX;
				_castHitY[cntr] = playerY + distance * stepY;
				if (wallHitX)
					_castSlice[cntr] = _castHitX[cntr];
				else
					_castSlice[cntr] = _castHitY[cntr];
				
				// make sure we have next angle
				angle += _sliceSize;
			}
		}

// ******************************************
// ***** Functions - Display Generation *****
// ******************************************

		/**
		* While this is intended as a test function, it could be useful as an
		* overlay in the automap. Simply put, this does the ray projection but
		* instead of generating a 3D view, it instead generates an alpha 
		* overlay showing on the map where the rays are projecting to.
		*/
		public function renderVisibility(
				playerX:Number,
				playerY:Number,
				playerD:Number,
				tileSize:int = 32)
		{
			var startTileX:int = int(playerX);
			var startTileY:int = int(playerY);
			var startAngle = (playerD - 45) * _sliceSize;
			var endAngle = (playerD + 45) * _sliceSize;
			performCast(playerX, playerY,playerD);
			graphics.clear();
			graphics.lineStyle(3, 0xFF00FF, 0.5);
			for (var cntr:int = 0; cntr < _displayWidth; ++cntr){
				graphics.moveTo(playerX * tileSize, playerY * tileSize);
//				graphics.lineTo(_castTileX[cntr] * tileSize, _castTileY[cntr] * tileSize);
				graphics.lineTo(_castHitX[cntr] * tileSize, _castHitY[cntr] * tileSize);				
			}

			graphics.lineStyle(3, 0x0000FF, 0.5);
			graphics.moveTo(playerX * tileSize, playerY * tileSize);
			graphics.lineTo(Math.sin(startAngle)*100+playerX*tileSize,
						-Math.cos(startAngle)*100+playerY*tileSize);
			graphics.moveTo(playerX * tileSize, playerY * tileSize);
			graphics.lineTo(Math.sin(endAngle)*100+playerX*tileSize,
						-Math.cos(endAngle)*100+playerY*tileSize);
		}

		/**
		 * Performs the rendering of the map.
		 * 
		 * @param playerX player x location on map
		 * @param playerY player y location on map
		 * @param playerD Direction the player is facing
		 */
		public function renderSolid( 
					playerX:Number,
					playerY:Number,
					playerD:Number)
		{
			var hght:Number;
			var half:Number;
			var colour:uint = 0xFF808080;
			performCast(playerX, playerY,playerD);
			_bitmapData.lock();
			_bitmapData.fillRect(_rect, 0x00FFFFFF);
			
			for (var cntr:int = 0; cntr < _displayWidth; ++cntr){
				hght = _castHeightTable[cntr] / (.000001 + _castDistance[cntr]);
				paintStrip(cntr, hght, _castTexture[cntr], _castSlice[cntr]);
			}
			_bitmapData.unlock();
		}

		/**
		 * This draws a strip.
		 *  
		 * @param x column to draw strip into
		 * @param height height of the strip
		 * @param strip the strip to draw (fraction is used to determine strip)
		 */
		public function paintStrip(x:int, h:int, texture:BitmapData, strip:Number):void
		{
			var strt:int, end:int, offset:int;
	
			if (h < _displayHeight) {
				offset = (_displayHeight - h) / 2;
				strt = 0;
				end = h;
			} else {
				offset = 0;
				strt = (h - _displayHeight) / 2;
				end = strt + _displayHeight;
			}
	        
			var stripPart:Number = strip - (Math.floor(strip));
			var stripX:int = texture.width * stripPart;
			var colorByte:uint = uint(255 * stripPart);
			var colour:uint = 0xff000000 | (colorByte << 16) | (colorByte << 8) | colorByte;
			var texelSize:Number =  texture.height / h;
	        for (var cntr:int = strt; cntr < end; ++cntr)
	        {
				_bitmapData.setPixel32(x,offset, //colour);
						texture.getPixel32(stripX, cntr * texelSize));
				++offset;
	        }
		}

	} // END CLASS	
} // END PACKAGE
