﻿/*
One of those Weeks Episode 45: Toronto?
Copyright (c) 2008,2012 Blazing Games Inc.
Code written by Billy D. Spelchan

This program is free software; you can redistribute it and/or modify it under 
the terms of the GNU General Public License as published by the Free Software 
Foundation; either version 2 of the License, or (at your option) any later 
version.

This program is distributed in the hope that it will be useful, but WITHOUT ANY
WARRANTY; without even the implied warranty of MERCHANTABILITY or FITNESS FOR A
PARTICULAR PURPOSE.  See the GNU General Public License for more details.

You should have received a copy of the GNU General Public License along with
this program; if not, write to the Free Software Foundation, Inc.,
59 Temple Place, Suite 330, Boston, MA  02111-1307  USA
*/
package com.blazinggames.ootw.ep45
{
	// Adobe Imports
	import flash.display.Graphics;
	import flash.display.MovieClip;

	/**
	 * This holds the maze data. As there is only one maze in episode 45,
	 * this class hard-codes the maze data (text data from my cqcs utility)
	 * into the class but it could easily be modified to load data.
	 * 
	 * @author Billy D. Spelchan
	 */
	public class Maze
	extends MovieClip
	{
// *********************
// ***** CONSTANTS *****
// *********************

		public const TILES:String = "-klno";
		public const TILE_EMPTY = 0;
		public const TILE_LADDER_START = 1;
		public const TILE_LADDER_END = 2;
		public const TILE_WALL = 3;
		public const TILE_ALSOWALL = 4;
		public const TILE_UNKNOWN = 5;
		public const RAW_MAP:String =
"nnnnnnnnnnnnnnnooonnnnnnnnnnnnnn"+
"n------------oooooooo----------n"+
"n-n-n-nnnnnn----------oooooooo-n"+
"n-n-n-n----nnn-nnnnnn-o--------n"+
"n-n-n---nn---n-n------o-ooooooon"+
"n-n-nnnnnnnn-n-n-nnnn-o--------n"+
"n-n-nn-------n-n-o--n-oooooooo-n"+
"n-n-nn-nnnnnnn-n---on--------o-n"+
"n-n-nn-n----nn-nno-onno-on-n-o-n"+
"n-n-nn-n-nn-nn---o-o--o-o--n-o-n"+
"n-n-------n-nn-n-n-oo-o-o-nn-o-n"+
"n-nnnnn-n-n-nn-n-n---------n-o-n"+
"n-n-----n-n--nln-nnnnnnnnn-n---n"+
"n-n-nnnnn-nnnnnn-n-------n-n-o-n"+
"n---n-----nn--n--n-nnnnn-n-n-o-n"+
"nnnnn-nnn----nnnnn-n---n-----o-n"+
"n---------nnnnkn-n-nnn-nnnn-oo-n"+
"n-nnnnnnnnn-nn-n-n-----n--n-oo-n"+
"n---n-nn--n------nnnnnnnn-n-o--n"+
"n-n-n--n-nnnnn-nnnnnnnn---n-o-on"+
"n-n-n-nn--------------n-n-n-o-on"+
"n-n-n-n--n-nnnnnnnnnn-n-n-n-o-on"+
"n-n-n---nn-n-n--nnn---n-n-n----n"+
"n-n-nnn--n-n---nn---n-n-n-n-nn-n"+
"n-n-nn--nn-nnn-n--n-n-n-n-n--n-n"+
"n-n----nn--n-----nn-n-n-n-nn-n-n"+
"n-nnnn-nn-nn-nn-nn--n-n-nnnn-n-n"+
"n-nn-n-nn-----n-n--nn-n---nn---n"+
"n----n--n-nnn-n-nnnn---nn-nnnn-n"+
"n-nnnnnnn-n---n------n---------n"+
"n---------nnnnnnnnnnnnnnnnnnnnnn"+
"nnnnnnnnnnnnnnnnnnnnnnnnnnnnnnnn";


		protected const AUTOMAPCOLORS:Array = new Array(
			0xffffff, 0xff0000, 0xff0000,0,0,0x808080
		);
		public const PLAYERSTARTX:Number = 14.5;
		public const PLAYERSTARTY:Number = 17.5;
		public const PLAYERSTARTD:Number = 0;

		public const STATE_LOSE:int = -1;
		public const STATE_NORMAL:int = 0;
		public const STATE_WIN:int = 1;
		
		public const DEGREEINRADS:Number = Math.PI / 180;

// *********************
// ***** VARIABLES *****
// *********************

		// for map management
		private var _map:Array;
		private var _automap:Array;
		
		// for player management
		var _player_x:Number;
		var _player_y:Number;
		var _player_dir:Number;

		// _render_movie is part of movie clip
		// _automap_movie is part of movie clip

// ***********************
// ***** CONSTRUCTOR *****
// ***********************

		public function Maze()
		{
			var cntr:int;
			
			// build the map
			_map = new Array(1024);
			_automap = new Array(1024);
			for (cntr = 0; cntr < 1024; ++cntr)
				_map[cntr] = TILES.indexOf(RAW_MAP.charAt(cntr));
			_render_movie.setMap(_map, 32);
			resetMap(true);
		}

// **************************************
// ***** Functions - Map Management *****
// **************************************

		public function getTile(x:uint, y:uint):int
		{
			var mapX = x & 31;
			var mapY = y & 31;
			return _map[(mapY*32) + mapX];
		}

		public function resetMap(clearAuto:Boolean = false)
		{
			_player_x = PLAYERSTARTX;
			_player_y = PLAYERSTARTY;
			_player_dir = PLAYERSTARTD;
			_render_movie.renderSolid(_player_x, _player_y, _player_dir);
			if (clearAuto) {
				for (var cntr:int = 0; cntr < 1024; ++cntr)
					_automap[cntr] = TILE_UNKNOWN;
			}
			updateAutomap();
		}

		public function generateAutomap()
		{
			var tileIndex:int = 0;
			var g:Graphics = _automap_movie.graphics;
			g.clear();
			for (var cntrY:int = 0; cntrY < 32; ++cntrY) {
				for (var cntrX:int = 0; cntrX < 32; ++cntrX) {
					g.beginFill(AUTOMAPCOLORS[_automap[tileIndex]]);
					g.drawRect(cntrX*4, cntrY*4, 4, 4);
					g.endFill();
					++tileIndex;
				}
			}
			g.beginFill(0xff00ff);
			g.drawRect(_player_x * 4 - 2, _player_y * 4 -2, 5, 5);
			g.endFill();
		}
		
		/**
		* This is tricky as we want to keep traps secret until they are sprung
		* but keep the traps visible if the player keeps the map. This is why
		* we force the player's location tile onto the automap but only change
		* non-set tiles and even then treat the trap tile as a space.
		*/
		public function updateAutomap()
		{
			var tileIndex:int;
			var tx:int, ty:int;
			for (var cntrX:int = -1; cntrX < 2;++cntrX) {
				tx = int(_player_x) + cntrX;
				for (var cntrY:int = -1; cntrY < 2; ++cntrY) {
					ty = int(_player_y) + cntrY;
					tileIndex = ty * 32 + tx;
					if (_automap[tileIndex] == TILE_UNKNOWN)
						_automap[tileIndex] = _map[tileIndex];
				}
			}
			tileIndex = int(_player_y) * 32 + int(_player_x);
			_automap[tileIndex] = _map[tileIndex];
			generateAutomap();
		}

// *****************************************
// ***** Functions - Player Management *****
// *****************************************

		public function moveForward(dist:Number):int
		{
			var angle:Number = _player_dir * Math.PI/180;
			var nx:Number = _player_x + dist * Math.sin(angle);
			var ny:Number = _player_y - dist * Math.cos(angle);
			var nt:int = getTile(uint(nx), uint(ny));
			if (nt < TILE_WALL) {
				var chck:Boolean = 
						( (Math.floor(nx) != Math.floor(_player_x)) ||
						(Math.floor(ny) != Math.floor(_player_y)));
				_player_x = nx;
				_player_y = ny;
				if (chck)
					updateAutomap();
			}
			_render_movie.renderSolid(_player_x, _player_y, _player_dir);
			if (nt == TILE_LADDER_END)
				return STATE_WIN;
			else if (nt == TILE_LADDER_START)
				return STATE_LOSE;
			else
				return STATE_NORMAL
		}
		
		public function turnLeft(dist:Number)
		{
			_player_dir -= (dist*90);
			_render_movie.renderSolid(_player_x, _player_y, _player_dir);
		}
		
		public function turnRight(dist:Number)
		{
			_player_dir += (dist*90);
			_render_movie.renderSolid(_player_x, _player_y, _player_dir);
		}
	} // END CLASS
} // END PACKAGE
